from __future__ import annotations

from copy import deepcopy
from typing import Any

from linflex import Vec2
from typing_extensions import Self


class Transform:  # Component (mixin class)
    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        instance = super().__new__(cls, *args, **kwargs)
        if (class_position := getattr(instance, "position", None)) is not None:
            instance.position = deepcopy(class_position)
        else:
            instance.position = Vec2.ZERO
        return instance

    position: Vec2
    rotation: float = 0
    top_level: bool = False

    # TODO: would be nice to figure out @overload with this function
    def with_position(
        self,
        position: Vec2 | None = None,
        /,
        x: float | None = None,
        y: float | None = None,
    ) -> Self:
        if position is None and x is None and y is None:
            raise TypeError(f"not all arguments can be {None} at the same time")
        if position is not None and (x is not None or y is not None):
            raise TypeError(
                "chose either positional argument 'position' "
                "or keyword arguments 'x' and/or 'y', not all three"
            )
        if position is not None:
            self.position = position
        if x is not None:
            self.position.x = x
        if y is not None:
            self.position.y = y
        return self

    # TODO: would be nice to figure out @overload with this function
    def with_global_position(
        self,
        global_position: Vec2 | None = None,
        /,
        x: float | None = None,
        y: float | None = None,
    ) -> Self:
        if global_position is None and x is None and y is None:
            raise TypeError(f"not all arguments can be {None} at the same time")
        if global_position is not None and (x is not None or y is not None):
            raise TypeError(
                "chose either positional argument 'global_position' "
                "or keyword arguments 'x' and/or 'y', not all three"
            )
        if global_position is not None:
            self.global_position = global_position
        if x is not None:
            self.set_global_x(x)
        if y is not None:
            self.set_global_y(y)
        return self

    def with_rotation(self, rotation: float, /) -> Self:
        self.rotation = rotation
        return self

    def with_global_rotation(self, global_rotation: float, /) -> Self:
        self.global_rotation = global_rotation
        return self

    def with_top_level(self, state: bool = True, /) -> Self:
        self.top_level = state
        return self

    def set_global_x(self, x: float, /) -> None:
        diff_x = x - self.global_position.x
        self.position.x += diff_x

    def set_global_y(self, y: float, /) -> None:
        diff_y = y - self.global_position.y
        self.position.y += diff_y

    @property
    def global_position(self) -> Vec2:
        """Returns a copy of the node's global position (in world space)

        `NOTE`: Cannot do `self.global_position.x += 5`,
        use `self.position += 5` instead, as it only adds a relative value

        `NOTE`: Cannot do `self.global_position.x = 42`,
        use `self.set_global_x(42)`

        Returns:
            Vec2: copy of global position
        """
        if self.top_level:
            return self.position.copy()
        global_position = self.position.copy()
        parent = self.parent  # type: ignore
        while isinstance(parent, Transform):
            # check for rotation, since cos(0) and sin(0) produces *approximate* values
            if parent.rotation:
                global_position = parent.position + global_position.rotated(
                    parent.rotation
                )
            else:
                global_position += parent.position
            if parent.top_level:
                return global_position
            parent = parent.parent  # type: ignore
        return global_position

    @global_position.setter
    def global_position(self, position: Vec2) -> None:
        """Sets the node's global position (world space)"""
        diff = position - self.global_position
        self.position += diff

    @property
    def global_rotation(self) -> float:
        """Computes the node's global rotation (world space)

        Returns:
            float: global rotation in radians
        """
        if self.top_level:
            return self.rotation
        global_rotation = self.rotation
        parent = self.parent  # type: ignore
        while isinstance(parent, Transform):
            global_rotation += parent.rotation
            if parent.top_level:
                return global_rotation
            parent = parent.parent  # type: ignore
        return global_rotation

    @global_rotation.setter
    def global_rotation(self, rotation: float) -> None:
        """Sets the node's global rotation (world space)"""
        diff = rotation - self.global_rotation
        self.rotation += diff



"""
Custom Annotations for Charz
============================

This file contains private annotations used across this package.

Whenever there is a "?" comment,
it means a type may or may not implement that field or mixin class.
"""

from __future__ import annotations as _annotations


from typing import (
    TypeVar as _TypeVar,
    Protocol as _Protocol,
    ClassVar as _ClassVar,
    runtime_checkable as _runtime_checkable,
    TYPE_CHECKING as _TYPE_CHECKING,
)

from linflex import (
    Vec2 as _Vec2,
    Vec2i as _Vec2i,
)
from colex import ColorValue as _ColorValue
from typing_extensions import Self as _Self

if _TYPE_CHECKING:
    from ._clock import DeltaClock as _DeltaClock
    from ._screen import Screen as _Screen
    from ._animation import (
        Animation as _Animation,
        AnimationSet as _AnimationMapping,
    )
    from ._components._collision import Hitbox as _Hitbox

T = _TypeVar("T")
_T_contra = _TypeVar("_T_contra", contravariant=True)


@_runtime_checkable
class FileLike(_Protocol[_T_contra]):
    def write(self, stream: _T_contra, /) -> object: ...
    def flush(self, /) -> None: ...
    def fileno(self, /) -> int: ...


@_runtime_checkable
class Engine(_Protocol):
    fps: float
    clock: _DeltaClock
    screen: _Screen
    is_running: bool


@_runtime_checkable
class Node(_Protocol):
    node_instances: _ClassVar[dict[int, Node]]
    uid: int

    def __init__(self) -> None: ...
    def with_parent(self, parent: Node | None, /) -> _Self: ...
    def with_process_priority(self, process_priority: int, /) -> _Self: ...
    def update(self, delta: float) -> None: ...
    def queue_free(self) -> None: ...
    def _free(self) -> None: ...


class TransformComponent(_Protocol):
    position: _Vec2
    rotation: float
    top_level: bool

    def with_position(
        self,
        position: _Vec2 | None = None,
        /,
        x: float | None = None,
        y: float | None = None,
    ) -> _Self: ...
    def with_global_position(
        self,
        global_position: _Vec2 | None = None,
        /,
        x: float | None = None,
        y: float | None = None,
    ) -> _Self: ...
    def with_rotation(self, rotation: float, /) -> _Self: ...
    def with_global_rotation(self, global_rotation: float, /) -> _Self: ...
    def with_top_level(self, state: bool = True, /) -> _Self: ...
    @property
    def global_position(self) -> _Vec2: ...
    @global_position.setter
    def global_position(self, position: _Vec2) -> None: ...
    @property
    def global_rotation(self) -> float: ...
    @global_rotation.setter
    def global_rotation(self, rotation: float) -> None: ...


@_runtime_checkable
class TransformNode(
    TransformComponent,
    Node,
    _Protocol,
): ...


class TextureComponent(_Protocol):
    texture_instances: _ClassVar[dict[int, TextureNode]]
    texture: list[str]
    visible: bool
    centered: bool
    z_index: int
    transparency: str | None

    def with_texture(self, texture_or_line: list[str] | str, /) -> _Self: ...
    def with_visibility(self, state: bool = True, /) -> _Self: ...
    def with_centering(self, state: bool = True, /) -> _Self: ...
    def with_z_index(self, z_index: int, /) -> _Self: ...
    def with_transparency(self, char: str | None, /) -> _Self: ...
    def hide(self) -> None: ...
    def show(self) -> None: ...
    def is_globally_visible(self) -> bool: ...
    @property
    def texture_size(self) -> _Vec2i: ...


@_runtime_checkable
class TextureNode(
    TextureComponent,
    TransformComponent,
    Node,
    _Protocol,
): ...


class ColorComponent(_Protocol):
    color: _ColorValue | None

    def with_color(
        self,
        color: _ColorValue | None,
        /,
    ) -> _Self: ...


@_runtime_checkable
class ColorNode(
    ColorComponent,
    TextureComponent,
    TransformNode,
    Node,
    _Protocol,
): ...


@_runtime_checkable
class Renderable(
    # `ColorComponent`?
    TextureComponent,
    TransformComponent,
    Node,
    _Protocol,
): ...


class AnimatedComponent(_Protocol):
    animations: _AnimationMapping
    current_animation: _Animation | None = None

    def with_animations(self, **animations: _Animation) -> _Self: ...
    def with_animation(
        self,
        animation_name: str,
        animation: _Animation,
        /,
    ) -> _Self: ...
    def add_animation(
        self,
        animation_name: str,
        animation: _Animation,
    ) -> None: ...
    def play(self, animation_name: str, /) -> None: ...
    def play_backwards(self, animation_name: str, /) -> None: ...
    def _wrapped_update_animated(self, _delta: float) -> None: ...


@_runtime_checkable
class AnimatedNode(
    # `ColorComponent`?
    AnimatedComponent,
    TextureComponent,
    TransformComponent,
    Node,
    _Protocol,
): ...


class ColliderComponent(_Protocol):
    collider_instances: _ClassVar[dict[int, ColliderNode]]
    hitbox: _Hitbox

    def with_hitbox(self, hitbox: _Hitbox, /) -> _Self: ...
    def get_colliders(self) -> list[ColliderNode]: ...
    def is_colliding_with(self, colldier_node: ColliderNode, /) -> bool: ...
    def is_colliding(self) -> bool: ...


@_runtime_checkable
class ColliderNode(
    # `ColorComponent`?
    # `TextureComponent`?
    ColliderComponent,
    TransformComponent,
    Node,
    _Protocol,
): ...



from __future__ import annotations

__all__ = ("SimpleMovement",)

from typing import TYPE_CHECKING, NoReturn

from linflex import Vec2

from ._transform import Transform

if TYPE_CHECKING:
    import keyboard
else:
    keyboard = None


def __getattr__(name: str) -> type[SimpleMovement] | NoReturn:
    try:
        global keyboard  # noqa: PLW0603
        if keyboard is None:
            import keyboard as _keyboard

            keyboard = _keyboard
    except ModuleNotFoundError as error:
        raise ModuleNotFoundError(
            "module 'keyboard' was not found,"
            " use 'charz' with 'keyboard' or 'all' feature flag,"
            " like depending on 'charz[keyboard]' in 'pyproject.toml'"
        ) from error
    if name not in __all__:
        raise AttributeError(f"module '{__name__}' has no attribute '{name}'")
    return SimpleMovement


class SimpleMovement:  # Component (mixin class)
    speed: float = 16

    def is_moving_left(self) -> bool:
        return keyboard.is_pressed("a")

    def is_moving_right(self) -> bool:
        return keyboard.is_pressed("d")

    def is_moving_up(self) -> bool:
        return keyboard.is_pressed("w")

    def is_moving_down(self) -> bool:
        return keyboard.is_pressed("s")

    def get_movement_direction(self) -> Vec2:
        return Vec2(
            self.is_moving_right() - self.is_moving_left(),
            self.is_moving_down() - self.is_moving_up(),
        )

    def get_movement_direction_strengths(self) -> Vec2:
        return Vec2(
            self.is_moving_right() - self.is_moving_left(),
            self.is_moving_down() - self.is_moving_up(),
        )

    # TODO: add automatic wrapping of this function, so no conflict with user `.update`
    def update(self, delta: float) -> None:
        super().update(delta)  # type: ignore
        assert isinstance(self, Transform), "Missing `Transform` mixin"
        self.position += self.get_movement_direction() * self.speed * delta



from __future__ import annotations

from colex import ColorValue
from typing_extensions import Self


class Color:  # Component (mixin class)
    color: ColorValue | None = None

    def with_color(self, color: ColorValue | None, /) -> Self:
        self.color = color
        return self



from __future__ import annotations

from dataclasses import dataclass
from copy import deepcopy
from typing import ClassVar, Any

from linflex import Vec2
from typing_extensions import Self

from .._components._transform import Transform
from .._annotations import ColliderNode


@dataclass(kw_only=True)
class Hitbox:
    size: Vec2
    centered: bool = False


class Collider:  # Component (mixin class)
    collider_instances: ClassVar[dict[int, ColliderNode]] = {}

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        instance = super().__new__(cls, *args, **kwargs)
        Collider.collider_instances[instance.uid] = instance  # type: ignore
        if (class_hitbox := getattr(instance, "hitbox", None)) is not None:
            instance.hitbox = deepcopy(class_hitbox)
        else:
            instance.hitbox = Hitbox(size=Vec2.ZERO)
        return instance

    hitbox: Hitbox

    def with_hitbox(self, hitbox: Hitbox, /) -> Self:
        self.hitbox = hitbox
        return self

    def get_colliders(self) -> list[Collider]:
        assert isinstance(self, ColliderNode)
        colliders: list[Collider] = []
        for node in Transform.transform_instances.values():
            if self is node:
                continue
            # NOTE: might swap who the `.is_colliding_with(...)` is checked on
            if isinstance(node, Collider) and node.is_colliding_with(self):
                colliders.append(node)
        return colliders

    def is_colliding_with(self, colldier_node: ColliderNode, /) -> bool:
        # TODO: consider `.global_rotation`
        assert isinstance(self, ColliderNode)
        start = self.global_position
        end = self.global_position + self.hitbox.size
        if self.hitbox.centered:
            start -= self.hitbox.size / 2
            end -= self.hitbox.size / 2
        return start <= colldier_node.global_position < end

    def is_colliding(self) -> bool:
        return bool(self.get_colliders())

    def _free(self) -> None:
        del Collider.collider_instances[self.uid]  # type: ignore
        super()._free()  # type: ignore



"""
Charz
=====

An object oriented terminal game engine

Includes
--------

- Annotations (from package `colex`)
  - `ColorValue`
- Math (from package `linflex`)
  - `lerp`
  - `sign`
  - `clamp`
  - `Vec2`
  - `Vec2i`
  - `Vec3`
- Modules
  - `text`
    - `fill`
    - `flip_h`
    - `flip_v`
    - `fill_lines`
    - `flip_lines_h`
    - `flip_lines_v`
    - `rotate`
  - `colex`    (dependency)
  - `keyboard` (optional dependency)
- Framework
  - `Engine`
  - `Clock`
  - `DeltaClock`
  - `Screen`
- Datastructures
  - `Animation`
  - `AnimationSet`
  - `Hitbox`
- Functions
  - `load_texture`
- Components
  - `Transform`
  - `Texture`
  - `Color`
  - `Animated`
  - `Collider`
- Nodes
  - `Camera`
  - `Node`
  - `Node2D`
  - `Sprite`
  - `Label`
  - `AnimatedSprite`
"""

__all__ = [
    "Engine",
    "Clock",
    "DeltaClock",
    "Screen",
    "Camera",
    "Node",
    "Node2D",
    "Transform",
    "lerp",
    "sign",
    "clamp",
    "Vec2",
    "Vec2i",
    "Vec3",
    "load_texture",
    "Texture",
    "Color",
    "ColorValue",
    "Label",
    "Sprite",
    "Animated",
    "AnimatedSprite",
    "Animation",
    "AnimationSet",
    "Collider",
    "Hitbox",
    "SimpleMovement",
    "AssetLoader",
    "text",
]

from typing import (
    TYPE_CHECKING as _TYPE_CHECKING,
    Literal as _Literal,
)

# re-exports
from linflex import lerp, sign, clamp, Vec2, Vec2i, Vec3
from colex import ColorValue

# exports
from ._engine import Engine
from ._clock import Clock, DeltaClock
from ._screen import Screen
from ._camera import Camera
from ._node import Node
from ._animation import Animation, AnimationSet
from ._components._transform import Transform
from ._components._texture import load_texture, Texture
from ._components._color import Color
from ._components._animated import Animated
from ._components._collision import Collider, Hitbox
from ._prefabs._node2d import Node2D
from ._prefabs._label import Label
from ._prefabs._sprite import Sprite
from ._prefabs._animated_sprite import AnimatedSprite
from ._asset_loader import AssetLoader
from . import text


# provide correct completion help - even if the required feature is not active
if _TYPE_CHECKING:
    from ._components._simple_movement import SimpleMovement

# lazy exports
# NOTE: add to `Literal` and tuple when adding new export
_lazy_objects: tuple[_Literal["SimpleMovement"]] = ("SimpleMovement",)
_loaded_objects: dict[str, object] = {
    name: obj
    for name, obj in globals().items()
    if name in __all__ and name not in _lazy_objects
}


# lazy load to properly load optional dependencies along the standard exports
# TODO: check if there is a step here that can be skipped
def __getattr__(name: str) -> object:
    if name in _loaded_objects:
        return _loaded_objects[name]
    elif name in _lazy_objects:
        # NOTE: manually add each branch
        if name == "SimpleMovement":
            from ._components._simple_movement import SimpleMovement

            _loaded_objects[name] = SimpleMovement
            return _loaded_objects[name]
        else:
            raise NotImplementedError(f"branch not implemented for '{name}'")
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")



from __future__ import annotations

from types import SimpleNamespace
from functools import partial
from pathlib import Path
from copy import deepcopy

from typing_extensions import Self

from ._components._texture import load_texture
from ._asset_loader import AssetLoader
from . import text


class AnimationClassProperties(type):
    _folder_path: Path = Path.cwd()

    # NOTE: This has to be set before importing local files in your project:
    # from charz importing ..., Animation, ...
    # Animation.folder_path = "src/animations"
    # from .local_file importing ...

    @property
    def folder_path(cls) -> Path:
        return cls._folder_path

    @folder_path.setter
    def folder_path(cls, new_path: Path | str) -> None:
        cls._folder_path = Path(new_path)
        if not cls._folder_path.exists():
            raise ValueError("invalid animation root folder path")


class Animation(metaclass=AnimationClassProperties):
    __slots__ = ("frames",)

    @classmethod
    def from_frames(
        cls,
        frames: list[list[str]],
        /,
        *,
        reverse: bool = False,
        flip_h: bool = False,
        flip_v: bool = False,
        fill: bool = True,
        fill_char: str = " ",
        unique: bool = True,
    ) -> Self:
        instance = super().__new__(cls)  # omit calling `__init__`
        # the negated parameters creates unique list instances,
        # so only copy if they are not present and `unique` is true,
        # else it would be copying an extra time for no reason
        if unique and not reverse and not flip_h and not flip_v and not fill:
            generator = deepcopy(frames)
        else:
            generator = frames

        if fill:  # NOTE: this fill logic has to be before flipping
            generator = map(partial(text.fill_lines, fill_char=fill_char), generator)
        if flip_h:
            generator = map(text.flip_lines_h, generator)
        if flip_v:
            generator = map(text.flip_lines_v, generator)
        if reverse:
            generator = reversed(list(generator))
        instance.frames = list(generator)
        return instance

    def __init__(
        self,
        animation_path: Path | str,
        /,
        *,
        reverse: bool = False,
        flip_h: bool = False,
        flip_v: bool = False,
        fill: bool = True,
        fill_char: str = " ",
    ) -> None:
        """Loads an `Animation` given a path to the folder where the animation is stored

        Args:
            folder_path (Path | str): path to folder where animation frames are stored as files.
            flip_h (bool, optional): flip frames horizontally. Defaults to False.
            flip_v (bool, optional): flip frames vertically. Defaults to False.
            fill (bool, optional): fill in to make shape of frames rectangular. Defaults to True.
            fill_char (str, optional): string of length 1 to fill with. Defaults to " ".
        """  # noqa: E501
        frame_directory = (
            Path.cwd()
            .joinpath(AssetLoader.animation_root)
            .joinpath(animation_path)
            .iterdir()
        )
        generator = map(load_texture, frame_directory)
        if fill:  # NOTE: this fill logic has to be before flipping
            generator = map(partial(text.fill_lines, fill_char=fill_char), generator)
        if flip_h:
            generator = map(text.flip_lines_h, generator)
        if flip_v:
            generator = map(text.flip_lines_v, generator)
        if reverse:
            generator = reversed(list(generator))
        self.frames = list(generator)

    def __repr__(self) -> str:
        # should never be empty, but if the programmer did it, show empty frame count
        if not self.frames:
            return f"{self.__class__.__name__}(N/A)"
        longest = 0
        shortest = 0
        tallest = 0
        lowest = 0
        # these are used as temporary variables in loop
        local_longest = 0
        local_shortest = 0
        local_tallest = 0
        local_lowest = 0
        for frame in self.frames:
            # compare all time best against best results per iteration
            # allow empty frame and frame with empty lines
            if not frame or not any(frame):
                continue
            local_longest = len(max(frame, key=len))
            longest = max(local_longest, longest)
            local_tallest = len(frame)
            tallest = max(local_tallest, tallest)
            local_shortest = len(min(frame, key=len))
            shortest = min(local_shortest, shortest)
            local_lowest = min(local_lowest, shortest)
        return (
            self.__class__.__name__
            + "("
            + f"{len(self.frames)}"
            + f":{shortest}x{lowest}"
            + f"->{longest}x{tallest}"
            + ")"
        )


class AnimationSet(SimpleNamespace):
    def __init__(self, **animations: Animation) -> None:
        super().__init__(**animations)

    def __getattribute__(self, name: str) -> Animation:
        return super().__getattribute__(name)

    def __setattr__(self, name: str, value: Animation) -> None:
        return super().__setattr__(name, value)



from __future__ import annotations

from colex import ColorValue
from linflex import Vec2

from .._node import Node

from .._animation import AnimationSet
from .._components._animated import Animated
from ._sprite import Sprite


class AnimatedSprite(Animated, Sprite):
    def __init__(
        self,
        parent: Node | None = None,
        *,
        process_priority: int | None = None,
        position: Vec2 | None = None,
        rotation: float | None = None,
        top_level: bool | None = None,
        texture: list[str] | None = None,
        visible: bool | None = None,
        centered: bool | None = None,
        z_index: int | None = None,
        transparency: str | None = None,
        color: ColorValue | None = None,
        animations: AnimationSet | None = None,
    ) -> None:
        super().__init__(
            parent=parent,
            process_priority=process_priority,
            position=position,
            rotation=rotation,
            top_level=top_level,
            texture=texture,
            visible=visible,
            centered=centered,
            z_index=z_index,
            transparency=transparency,
            color=color,
        )
        if animations is not None:
            self.animations = animations



from __future__ import annotations

from enum import Flag, unique, auto

from linflex import Vec2
from typing_extensions import Self

from ._node import Node, NodeMixinSorter
from ._components._transform import Transform


@unique
class CameraMode(Flag):
    FIXED = auto()
    CENTERED = auto()
    INCLUDE_SIZE = auto()


class CameraClassAttributes(NodeMixinSorter):
    MODE_FIXED: CameraMode = CameraMode.FIXED
    MODE_CENTERED: CameraMode = CameraMode.CENTERED
    MODE_INCLUDE_SIZE: CameraMode = CameraMode.INCLUDE_SIZE
    _current: Camera

    @property
    def current(self) -> Camera:
        if not hasattr(self, "_current"):
            self._current = Camera()  # Create default camera if none exists
        return self._current

    @current.setter
    def current(self, new: Camera | None) -> None:
        if new is None:
            self._current = Camera()
        else:
            self._current = new


class Camera(Transform, Node, metaclass=CameraClassAttributes):
    mode: CameraMode = CameraMode.FIXED

    def __init__(
        self,
        parent: Node | None = None,
        *,
        process_priority: int | None = None,
        position: Vec2 | None = None,
        rotation: float | None = None,
        top_level: bool | None = None,
        mode: CameraMode | None = None,
        current: bool | None = False,
    ) -> None:
        if parent is not None:
            self.parent = parent
        if process_priority is not None:
            self.process_priority = process_priority
        if position is not None:
            self.position = position
        if rotation is not None:
            self.rotation = rotation
        if top_level is not None:
            self.top_level = top_level
        if mode is not None:
            self.mode = mode
        if current is not None:
            self.set_current()

    def set_current(self) -> None:
        Camera.current = self

    def as_current(self) -> Self:
        self.set_current()
        return self

    def is_current(self) -> bool:
        return Camera.current is self

    def with_mode(self, mode: CameraMode, /) -> Self:
        self.mode = mode
        return self



from __future__ import annotations

from itertools import count
from typing import Any, ClassVar

from typing_extensions import Self


class NodeMixinSorter(type):
    """Node metaclass for initializing `Node` subclass after other `mixin` classes"""

    def __new__(
        cls,
        name: str,
        bases: tuple[type, ...],
        attrs: dict[str, object],
    ) -> type:
        def sorter(base: type) -> bool:
            return isinstance(base, Node)

        sorted_bases = tuple(sorted(bases, key=sorter))
        new_type = super().__new__(cls, name, sorted_bases, attrs)
        return new_type


class Node(metaclass=NodeMixinSorter):
    _queued_nodes: ClassVar[list[Node]] = []
    _uid_counter: ClassVar[count] = count(0, 1)
    node_instances: ClassVar[dict[int, Node]] = {}

    def __new__(cls, *_args: Any, **_kwargs: Any) -> Self:
        # NOTE: additional args and kwargs are ignored!
        instance = super().__new__(cls)
        instance.uid = next(Node._uid_counter)
        Node.node_instances[instance.uid] = instance
        return instance

    uid: int  # is set in `Node.__new__`
    parent: Node | None = None
    process_priority: int = 0

    def __init__(self, parent: Node | None = None) -> None:
        if parent is not None:
            self.parent = parent

    def with_parent(self, parent: Node | None, /) -> Self:
        self.parent = parent
        return self

    def with_process_priority(self, process_priority: int, /) -> Self:
        self.process_priority = process_priority
        return self

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(#{self.uid})"

    def update(self, delta: float) -> None: ...

    def queue_free(self) -> None:
        if self not in Node._queued_nodes:
            Node._queued_nodes.append(self)

    def _free(self) -> None:
        del Node.node_instances[self.uid]



from __future__ import annotations

import time
from typing import Literal


class Clock:
    """`Clock` base class, without delta time"""

    _tps: float | None = None

    @property
    def tps(self) -> float | None:
        return self._tps

    @tps.setter
    def tps(self, value: float | None) -> None:
        self._tps = value

    @property
    def delta(self) -> Literal[0]:
        """Returns the delta time between ticks

        Returns:
            Literal[0]: always returns 0, since `Clock` class does not calculate delta time
        """  # noqa: E501
        return 0

    def tick(self) -> None:
        """Does nothing on its own. Exists for better coupling, when extending the `Clock` class"""  # noqa: E501
        return


class DeltaClock(Clock):
    """`DeltaClock` class, with delta time calculation"""

    def __init__(self) -> None:
        self._delta: float = 0
        self._target_delta: float = 0
        self._last_tick = time.perf_counter()

    @property
    def tps(self) -> float | None:
        return self._tps

    @tps.setter
    def tps(self, value: float | None) -> None:
        self._tps = value
        if value is not None:
            self._target_delta = 1.0 / value

    @property
    def delta(self) -> float:
        return self._delta

    def tick(self) -> None:
        """Sleeps for the remaining time to maintain desired `tps`"""
        current_time = time.perf_counter()
        elapsed_time = current_time - self._last_tick
        sleep_time = self._target_delta - elapsed_time
        if sleep_time > 0:
            time.sleep(sleep_time)
            self._last_tick = time.perf_counter()
        else:
            self._last_tick = current_time
        self._delta = max(0, sleep_time)



import colex  # Color constants and styling
import keyboard  # For taking key inputs
from charz import *


class Player(Sprite):
    SPEED: int = 4  # Defining local constant
    color = colex.RED  # In reality just a string, like "\x1b[31m" for red
    centered = True  # Apply sprite centereing - Handled by `charz`
    texture = [  # A texture may be defined as a class variable, of type `list[str]`
        "  O",
        "/ | \\",
        " / \\",
    ]

    def update(self, delta: float) -> None:  # This method is called every frame
        if keyboard.is_pressed("a"):
            self.position.y -= self.SPEED * delta
        if keyboard.is_pressed("d"):
            self.position.y += self.SPEED * delta
        if keyboard.is_pressed("s"):
            self.position.y += self.SPEED * delta
        if keyboard.is_pressed("w"):
            self.position.y -= self.SPEED * delta


class Game(Engine):
    fps = 12
    screen = Screen(auto_resize=True)
    clear_console = True

    def __init__(self) -> None:
        Camera.current.mode = Camera.MODE_CENTERED
        self.player = Player(position=Vec2(10, 5))

    def update(self, _delta: float) -> None:
        if keyboard.is_pressed("q"):
            self.is_running = False
        if keyboard.is_pressed("e"):
            self.player.queue_free()  # `Engine` will drop reference to player
            # NOTE: Player reference is still kept alive by `App`, but it won't be updated


if __name__ == "__main__":
    game = Game()
    game.run()



"""
Text utility module
===================

Utility for flipping characters/lines. Support for rotating characters

Includes
--------

- `fill`
- `flip_h`
- `flip_v`
- `fill_lines`
- `flip_lines_h`
- `flip_lines_v`
- `rotate`
"""

from __future__ import annotations as _annotations

from math import tau as _TAU


# IDEA: make TextTranslator (as in Transform props),
#       as public export,
#       and wrap methods of 1 default instance with module functions in this file


# predefined horizontal conversions
_h_conversions: dict[str, str] = {  # horizontal flip
    "/": "\\",
    "(": ")",
    "[": "]",
    "{": "}",
    ">": "<",
    "´": "`",
    "d": "b",
    "q": "p",
}
# mirroring `_h_conversions`
# fmt: off
_h_conversions.update({
    value: key
    for key, value in _h_conversions.items()
})
# unmirrored `_h_conversions` for monodirectional translations
_h_conversions.update({
    "7": "<"
})
# fmt: on
# predefined vertical conversions
_v_conversions: dict[str, str] = {  # vertical flip
    "/": "\\",
    ".": "'",
    ",": "`",
    "¯": "_",
    "b": "p",
    "q": "d",
    "w": "m",
    "W": "M",
    "v": "^",
    "V": "A",
}
# mirroring `_v_conversions`
# fmt: off
_v_conversions.update({
    value: key
    for key, value in _v_conversions.items()
})
# fmt: on
# unmirrored `_v_conversions` for monodirectional translations
# fmt: off
# _v_conversions.update({
#     # none for now...
# })
# fmt: on
# predefined rotational conversions
_r_conversions: dict[str, tuple[str, ...]] = {  # rotational
    "|": ("|", "\\", "-", "/"),
    ".": (".", "'"),
    "b": ("b", "p", "q", "d"),
    "9": ("9", "6"),
}
# (I marked vars with underscore as they are not for export, and this is a public module)
# creating mirrored {char: variants} pairs, for pairs already defined in `_r_conversions`
# mirror `_r_conversions` (adds variants as their own keys)
for _options in list(_r_conversions.values()):
    for _idx, _value in enumerate(_options):
        if _idx == 0:  # does not need to re-add the initial pair
            continue
        assert _value not in _r_conversions, "cannot add existing value: " + repr(_value)
        _before = _options[:_idx]
        _after = _options[_idx:]
        _new_values = (*_after, *_before)
        _r_conversions[_value] = _new_values
# unmirrored `_r_conversions` for spesific lookup translations
# fmt: off
# _r_conversions.update({
#     # none for now...
# })
# fmt: on


def fill(line: str, *, width: int, fill_char: str = " ") -> str:
    """Fills a single left-justified line with a string of length 1

    Args:
        line (str): line to be filled
        width (int): maximum width of output string
        fill_char (str, optional): string of length 1 to fill line with. Defaults to " ".

    Returns:
        str: line filled with fill character
    """
    return line.ljust(width, fill_char)


def flip_h(line: str, /) -> str:
    """Flips a single line horizontally. Also works with a single character

    Args:
        line (list[str]): content to ble flipped

    Returns:
        list[str]: flipped line or character
    """
    return "".join(_h_conversions.get(char, char) for char in reversed(line))


def flip_v(line: str, /) -> str:
    """Flips a single line vertically. Also works with a single character

    Args:
        line (list[str]): content to ble flipped

    Returns:
        list[str]: flipped line or character
    """
    return "".join(_v_conversions.get(char, char) for char in line)


def fill_lines(lines: list[str], *, fill_char: str = " ") -> list[str]:
    """Fill lines with fill character, based on longest line.
    Usefull for filling textures, so that it gets a nice rectangular shape.
    Good for centering and flipping textures.

    Args:
        lines (list[str]): lines to be filled
        fill_char (str, optional): string of length 1 to fill line with. Defaults to " ".

    Returns:
        list[str]: rectangular filled lines
    """
    if not any(lines):  # allow empty lines
        return []  # but still return unique list
    longest = len(max(lines, key=len))
    return [fill(line, width=longest, fill_char=fill_char) for line in lines]


def flip_lines_h(lines: list[str], /) -> list[str]:
    """Flips lines horizontally. Usefull for flipping textures

    Args:
        lines (list[str]): lines of strings or texture

    Returns:
        list[str]: flipped content
    """
    return [flip_h(line) for line in lines]


def flip_lines_v(lines: list[str], /) -> list[str]:
    """Flips lines vertically. Usefull for flipping textures

    Args:
        lines (list[str]): lines of strings or texture

    Returns:
        list[str]: flipped content
    """
    return [flip_v(line) for line in reversed(lines)]


def rotate(char: str, /, angle: float) -> str:
    """Returns symbol when rotated by angle counter clockwise

    Args:
        char (str): character to rotate
        angle (float): counter clockwise rotation in radians

    Returns:
        str: rotated character or original character
    """
    if char in _r_conversions:
        sector_count = len(_r_conversions[char])
        sector_rads = _TAU / sector_count
        half_sector_rads = sector_rads / 2
        total_rads = (angle + half_sector_rads) % _TAU
        index = int(total_rads / sector_rads) % sector_count
        return _r_conversions[char][index]
    return char



from __future__ import annotations

import os
import sys
from enum import Enum, unique, auto

from linflex import Vec2i
from colex import ColorValue, RESET

from ._camera import Camera, CameraMode
from ._components._transform import Transform
from ._components._texture import Texture
from ._annotations import FileLike, Renderable


CONSOLE_CLEAR_CODE = "\x1b[2J\x1b[H"
CURSOR_HIDE_CODE = "\x1b[?25l"
CURSOR_SHOW_CODE = "\x1b[?25h"


@unique
class ColorChoice(Enum):
    AUTO = auto()
    ALWAYS = auto()
    NEVER = auto()


class ScreenClassProperties(type):
    COLOR_CHOICE_AUTO = ColorChoice.AUTO
    COLOR_CHOICE_ALWAYS = ColorChoice.ALWAYS
    COLOR_CHOICE_NEVER = ColorChoice.NEVER


class Screen(metaclass=ScreenClassProperties):
    stream: FileLike[str] = sys.stdout  # default stream, may be redirected
    # screen texture buffer with (char, color) tuple
    buffer: list[list[tuple[str, ColorValue | None]]]

    def __init__(
        self,
        width: int = 16,
        height: int = 12,
        *,
        auto_resize: bool = False,
        initial_clear: bool = False,
        hide_cursor: bool = True,
        transparency_fill: str = " ",
        color_choice: ColorChoice = ColorChoice.AUTO,
        stream: FileLike[str] | None = None,
        margin_right: int = 1,
        margin_bottom: int = 1,
    ) -> None:
        self.width = width
        self.height = height
        self.color_choice = color_choice
        if stream is not None:
            self.stream = stream
            # NOTE: uses class variable `Screen.stream` by default
        self.margin_right = margin_right
        self.margin_bottom = margin_bottom
        self._auto_resize = auto_resize
        self.initial_clear = initial_clear
        self.hide_cursor = hide_cursor
        self._resize_if_necessary()
        self.transparency_fill = transparency_fill
        self.buffer = []
        self.clear()  # for populating the screen buffer
        self.startup()

    def startup(self) -> None:
        if self.is_using_ansi():
            if self.initial_clear:
                self.stream.write(CONSOLE_CLEAR_CODE)
                self.stream.flush()
            if self.hide_cursor:
                self.stream.write(CURSOR_HIDE_CODE)
                self.stream.flush()

    def cleanup(self) -> None:
        if self.hide_cursor and self.is_using_ansi():
            self.stream.write(CURSOR_SHOW_CODE)
            self.stream.flush()

    @property
    def auto_resize(self) -> bool:
        return self._auto_resize

    @auto_resize.setter
    def auto_resize(self, state: bool) -> None:
        self._auto_resize = state
        self._resize_if_necessary()

    def _resize_if_necessary(self) -> None:
        # NOTE: does not mutate screen buffer
        if self.auto_resize:
            try:
                fileno = self.stream.fileno()
            except (ValueError, OSError):
                # do not resize if not proper `.stream.fileno()` is available,
                # like `io.StringIO.fileno()`
                return
            try:
                terminal_size = os.get_terminal_size(fileno)
            except (ValueError, OSError):
                return
            self.width = terminal_size.columns - self.margin_right
            self.height = terminal_size.lines - self.margin_bottom

    @property
    def size(self) -> Vec2i:
        return Vec2i(self.width, self.height)

    @size.setter
    def size(self, value: Vec2i) -> None:
        width, height = value.to_tuple()
        if not isinstance(width, int):
            raise ValueError(f"width cannot be of type '{type(value)}', expected 'int'")
        if not isinstance(height, int):
            raise ValueError(f"height cannot be of type '{type(value)}', expected 'int'")
        self.width = width
        self.height = height
        self._resize_if_necessary()

    def is_using_ansi(self) -> bool:
        """Returns whether its using ANSI escape and color codes

        Checks first `.color_choice`. Returns `True` if set to `ALWAYS`,
        and `False` if set to `NEVER`. If set to `AUTO`, check whether a `tty` is detected

        Returns:
            bool: ansi use
        """
        if self.color_choice is ColorChoice.ALWAYS:
            return True
        try:
            fileno = self.stream.fileno()
        except (ValueError, OSError):
            is_a_tty = False
        else:
            try:
                is_a_tty = os.isatty(fileno)
            except OSError:
                is_a_tty = False
        # is not a tty or `ColorChoice.NEVER`
        return self.color_choice is ColorChoice.AUTO and is_a_tty

    def get_actual_size(self) -> Vec2i:
        try:
            fileno = self.stream.fileno()
        except (ValueError, OSError):
            return self.size.copy()
        try:
            terminal_size = os.get_terminal_size(fileno)
        except (ValueError, OSError):
            return self.size.copy()
        actual_width = min(self.width, terminal_size.columns - self.margin_right)
        actual_height = min(self.height, terminal_size.lines - self.margin_bottom)
        return Vec2i(actual_width, actual_height)

    def clear(self) -> None:
        self.buffer = [
            # (char, color) group
            [(self.transparency_fill, None) for _ in range(self.width)]
            for _ in range(self.height)
        ]

    def render(self, node: Renderable, /) -> None:  # noqa: C901
        if not node.is_globally_visible():  # skip if node is invisible
            return
        # current camera should never be None or other class than 'Camera',
        # or subclass of it
        if Camera.current is None or not isinstance(Camera.current, Camera):
            raise TypeError(
                "'Camera.current' cannot be of type "
                f"'{type(Camera.current)}' while rendering"
            )

        color: ColorValue | None = getattr(node, "color")  # noqa: B009
        # TODO: implement rotation when rendering
        # node_global_rotation = node.global_rotation
        node_global_position = node.global_position

        # determine whether to use use the parent of current camera
        # or its parent as anchor for viewport
        anchor = Camera.current
        if (
            not Camera.current.top_level
            and Camera.current.parent is not None
            and isinstance(Camera.current.parent, Transform)
        ):
            anchor = Camera.current.parent
        relative_position = node_global_position - anchor.global_position

        if Camera.current.mode & CameraMode.CENTERED:
            relative_position += self.size / 2

        # include half size of camera parent when including size
        viewport_global_position = Camera.current.global_position
        if (
            Camera.current.mode & CameraMode.INCLUDE_SIZE
            and Camera.current.parent is not None
            and isinstance(Camera.current.parent, Texture)
        ):
            # adds half of camera's parent's texture size
            # TODO: cache `.parent.texture_size` for the whole iteration in main loop
            viewport_global_position += Camera.current.parent.texture_size / 2

        actual_size = self.get_actual_size()

        texture_size = node.texture_size  # store as variable for performance
        x = int(relative_position.x)
        y = int(relative_position.y)
        if node.centered:
            x = int(relative_position.x - (texture_size.x / 2))
            y = int(relative_position.y - (texture_size.y / 2))

        # TODO: consider nodes with rotation
        # out of bounds
        if x + texture_size.x < 0 or x > actual_size.x:
            return
        if y + texture_size.y < 0 or y > actual_size.y:
            return

        for y_offset, line in enumerate(node.texture):
            y_final = y + y_offset
            for x_offset, char in enumerate(line):
                if char == node.transparency:  # skip transparent char
                    continue
                x_final = x + x_offset
                # insert char into screen buffer if visible
                if 0 <= x_final < actual_size.x and 0 <= y_final < actual_size.y:
                    self.buffer[y_final][x_final] = (char, color)
        # TODO: implement render with rotation

    def show(self) -> None:
        actual_size = self.get_actual_size()
        # construct frame
        out = ""
        is_using_ansi = self.is_using_ansi()
        for lino, row in enumerate(self.buffer[: actual_size.y], start=1):
            for char, color in row[: actual_size.x]:
                if is_using_ansi:
                    if color is None:
                        out += RESET + char
                    else:
                        out += RESET + color + char
                else:
                    out += char
            if lino != actual_size.y:  # not at end
                out += "\n"
        if is_using_ansi:
            out += RESET
            cursor_move_code = f"\x1b[{actual_size.y - 1}A" + "\r"
            out += cursor_move_code
        # write and flush
        self.stream.write(out)
        self.stream.flush()

    def refresh(self) -> None:
        self._resize_if_necessary()
        self.clear()
        for node in sorted(  # NOTE: iterator becomes a `list`
            # NOTE: `list` is faster than `tuple`, when copying
            list(Texture.texture_instances.values()),  # iterate a copy
            key=lambda node: node.z_index,
        ):
            self.render(node)
        self.show()



from __future__ import annotations

from pathlib import Path
from typing import Any, NoReturn, final


class AssetLoaderClassProperties(type):
    # class variables for configuartion
    _texture_root: Path = Path.cwd()
    _animation_root: Path = Path.cwd()

    # NOTE: Theese have to be set before importing local files in your project:
    # from charz importing ..., AssetLoader, ...
    # AssetLoader.texture_root = "src/sprites"
    # AssetLoader.animation_root = "src/animations"
    # from .local_file importing ...

    @property
    def texture_root(cls) -> Path:
        return cls._texture_root

    @texture_root.setter
    def texture_root(cls, new_path: Path | str) -> None:
        cls._texture_root = Path(new_path)
        if not cls._texture_root.exists():
            raise ValueError("invalid sprite root folder path")

    @property
    def animation_root(cls) -> Path:
        return cls._animation_root

    @animation_root.setter
    def animation_root(cls, new_path: Path | str) -> None:
        cls._animation_root = Path(new_path)
        if not cls.animation_root.exists():
            raise ValueError("invalid animation root folder path")


@final
class AssetLoader(metaclass=AssetLoaderClassProperties):
    # prevent instantiating, as this class only has class methods and class variables
    def __new__(cls, *_args: Any, **_kwargs: Any) -> NoReturn:
        raise TypeError(f"{cls.__name__} cannot be instantiated")



from __future__ import annotations

from typing import Any

from typing_extensions import Self

from ._clock import Clock, DeltaClock
from ._screen import Screen
from ._node import Node


class EngineMixinSorter(type):
    """Engine metaclass for initializing `Engine` subclass after other `mixin` classes"""

    def __new__(
        cls,
        name: str,
        bases: tuple[type, ...],
        attrs: dict[str, object],
    ) -> type:
        def sorter(base: type) -> bool:
            return isinstance(base, Engine)

        sorted_bases = tuple(sorted(bases, key=sorter))
        new_type = super().__new__(cls, name, sorted_bases, attrs)
        return new_type


class Engine(metaclass=EngineMixinSorter):
    fps: float | None = 16
    clock: Clock = DeltaClock()
    screen: Screen = Screen()
    is_running: bool = False

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        instance = super().__new__(cls, *args, **kwargs)
        # set `.clock.tps` with `.fps` set from class attribute
        instance.clock.tps = instance.fps
        return instance  # type: ignore

    def update(self, delta: float) -> None: ...

    def run(self) -> None:
        delta = self.clock.delta  # initial delta
        self.is_running = True

        while self.is_running:  # main loop
            self.update(delta)
            for queued_node in Node._queued_nodes:
                queued_node._free()
            Node._queued_nodes *= 0  # NOTE: faster way to do `.clear()`
            # NOTE: `list` is faster than `tuple`, when copying
            for node in list(Node.node_instances.values()):  # iterating copy
                node.update(delta)
            self.screen.refresh()
            self.clock.tick()
            delta = self.clock.delta



from __future__ import annotations

from copy import deepcopy
from pathlib import Path
from typing import Any, ClassVar

from linflex import Vec2i
from typing_extensions import Self

from .. import text
from .._asset_loader import AssetLoader
from .._annotations import TextureNode


# TODO: in future versions, add caching
def load_texture(
    texture_path: Path | str,
    /,
    *,
    flip_h: bool = False,
    flip_v: bool = False,
    fill: bool = True,
    fill_char: str = " ",
) -> list[str]:
    """Loads texture from file

    Args:
        texture_path (Path | str): path to file with texture.
        flip_h (bool, optional): flip horizontally. Defaults to False.
        flip_v (bool, optional): flip vertically. Defaults to False.
        fill (bool, optional): fill in to make shape rectangular. Defaults to True.
        fill_char (str, optional): string of length 1 to fill with. Defaults to " ".

    Returns:
        list[str]: loaded texture
    """
    # fmt: off
    file = (
        Path.cwd()
        .joinpath(AssetLoader.texture_root)
        .joinpath(texture_path)
    )
    # fmt: on
    content = file.read_text(encoding="utf-8")
    texture = content.splitlines()
    if fill:  # NOTE: this fill logic has to be before flipping
        texture = text.fill_lines(texture, fill_char=fill_char)
    if flip_h:
        texture = text.flip_lines_h(texture)
    if flip_v:
        texture = text.flip_lines_v(texture)
    return texture


class Texture:  # Component (mixin class)
    texture_instances: ClassVar[dict[int, TextureNode]] = {}

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        instance = super().__new__(cls, *args, **kwargs)
        Texture.texture_instances[instance.uid] = instance  # type: ignore
        if (class_texture := getattr(instance, "texture", None)) is not None:
            instance.texture = deepcopy(class_texture)
        else:
            instance.texture = []
        return instance

    texture: list[str]
    visible: bool = True
    centered: bool = False
    z_index: int = 0
    transparency: str | None = None

    def with_texture(self, texture_or_line: list[str] | str, /) -> Self:
        if isinstance(texture_or_line, str):
            self.texture = [texture_or_line]
            return self
        self.texture = texture_or_line
        return self

    def with_visibility(self, state: bool = True, /) -> Self:
        self.visible = state
        return self

    def with_centering(self, state: bool = True, /) -> Self:
        self.centered = state
        return self

    def with_z_index(self, z_index: int, /) -> Self:
        self.z_index = z_index
        return self

    def with_transparency(self, char: str | None, /) -> Self:
        self.transparency = char
        return self

    def hide(self) -> None:
        self.visible = False

    def show(self) -> None:
        self.visible = True

    def is_globally_visible(self) -> bool:  # global visibility
        """Checks whether the node and its ancestors are visible

        Returns:
            bool: global visibility
        """
        if not self.visible:
            return False
        parent = self.parent  # type: ignore
        while parent is not None:
            if not isinstance(parent, Texture):
                return True
            if not parent.visible:
                return False
            parent = parent.parent  # type: ignore
        return True

    @property
    def texture_size(self) -> Vec2i:
        """Get the size of the texture

        Computed in O(n*m), where n is the number of lines and m is the length of the longest line

        Returns:
            Vec2i: texture size
        """  # noqa: E501
        if not self.texture:
            return Vec2i.ZERO
        return Vec2i(
            len(max(self.texture, key=len)),  # size of longest line
            len(self.texture),  # line count
        )

    def _free(self) -> None:
        del Texture.texture_instances[self.uid]  # type: ignore
        super()._free()  # type: ignore



from __future__ import annotations

from colex import ColorValue
from linflex import Vec2
from typing_extensions import Self

from .._node import Node
from ._sprite import Sprite


class Label(Sprite):
    newline: str = "\n"
    tab_size: int = 4
    tab_char: str = "\t"
    tab_fill: str = " "

    def __init__(
        self,
        parent: Node | None = None,
        *,
        process_priority: int | None = None,
        position: Vec2 | None = None,
        rotation: float | None = None,
        top_level: bool | None = None,
        texture: list[str] | None = None,
        visible: bool | None = None,
        centered: bool | None = None,
        z_index: int | None = None,
        transparency: str | None = None,
        color: ColorValue | None = None,
        text: str | None = None,
        newline: str | None = None,
        tab_size: int | None = None,
        tab_char: str | None = None,
        tab_fill: str | None = None,
    ) -> None:
        super().__init__(
            parent=parent,
            process_priority=process_priority,
            position=position,
            rotation=rotation,
            top_level=top_level,
            texture=texture,
            visible=visible,
            centered=centered,
            z_index=z_index,
            transparency=transparency,
            color=color,
        )
        if text is not None:
            self.text = text
        if newline is not None:
            self.newline = newline
        if tab_size is not None:
            self.tab_size = tab_size
        if tab_char is not None:
            self.tab_char = tab_char
        if tab_fill is not None:
            self.tab_fill = tab_fill

    def with_newline(self, newline: str, /) -> Self:
        self.newline = newline
        return self

    def with_tab_size(self, tab_size: int, /) -> Self:
        self.tab_size = tab_size
        return self

    def with_tab_char(self, tab_char: str, /) -> Self:
        self.tab_char = tab_char
        return self

    def with_tab_fill(self, tab_fill: str, /) -> Self:
        self.tab_fill = tab_fill
        return self

    def with_text(self, text: str, /) -> Self:
        self.text = text
        return self

    @property
    def text(self) -> str:
        joined_lines = self.newline.join(self.texture)
        return joined_lines.replace(self.tab_fill * self.tab_size, self.tab_char)

    @text.setter
    def text(self, value: str) -> None:
        tab_replaced = self.newline.replace(self.tab_char, self.tab_fill * self.tab_size)
        self.texture = value.split(tab_replaced)



from __future__ import annotations as _annotations

from linflex import Vec2

from .._node import Node
from .._components._transform import Transform


class Node2D(Transform, Node):
    def __init__(
        self,
        parent: Node | None = None,
        *,
        process_priority: int | None = None,
        position: Vec2 | None = None,
        rotation: float | None = None,
        top_level: bool | None = None,
    ) -> None:
        super().__init__(parent=parent)
        if process_priority is not None:
            self.process_priority = process_priority
        if position is not None:
            self.position = position
        if rotation is not None:
            self.rotation = rotation
        if top_level:  # `bool` and `None` check combined
            self.top_level = top_level

    def __str__(self) -> str:
        return (
            self.__class__.__name__
            + "("
            + f"#{self.uid}"
            + f":{self.position}"
            + f":{self.rotation}"
            + ")"
        )



from __future__ import annotations

from colex import ColorValue
from linflex import Vec2

from .._node import Node
from .._components._texture import Texture
from .._components._color import Color
from ._node2d import Node2D


class Sprite(Color, Texture, Node2D):
    def __init__(
        self,
        parent: Node | None = None,
        *,
        process_priority: int | None = None,
        position: Vec2 | None = None,
        rotation: float | None = None,
        top_level: bool | None = None,
        texture: list[str] | None = None,
        visible: bool | None = None,
        centered: bool | None = None,
        z_index: int | None = None,
        transparency: str | None = None,
        color: ColorValue | None = None,
    ) -> None:
        super().__init__(
            parent=parent,
            process_priority=process_priority,
            position=position,
            rotation=rotation,
            top_level=top_level,
        )
        if texture is not None:
            self.texture = texture
        if visible is not None:
            self.visible = visible
        if centered is not None:
            self.centered = centered
        if z_index is not None:
            self.z_index = z_index
        if transparency is not None:
            self.transparency = transparency
        if color is not None:
            self.color = color

    def __str__(self) -> str:
        return (
            self.__class__.__name__
            + "("
            + f"#{self.uid}"
            + f":{self.position}"
            + f":{round(self.rotation, 2)}R"
            + f":{'{}x{}'.format(*self.texture_size.to_tuple())}"
            + f":{repr(self.color)}"
            + ")"
        )



from __future__ import annotations

from functools import wraps
from copy import deepcopy
from enum import Enum, unique, auto
from typing import Any

from typing_extensions import Self

from .._animation import AnimationSet, Animation
from .._annotations import AnimatedNode


@unique
class PlaybackDirection(Enum):
    FORAWRD = auto()
    BACKWARD = auto()


class Animated:  # Component (mixin class)
    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        instance = super().__new__(cls, *args, **kwargs)
        if (class_animations := getattr(instance, "animations", None)) is not None:
            instance.animations = deepcopy(class_animations)
        else:
            instance.animations = AnimationSet()

        # inject `._wrapped_update_animated()` into `.update()`
        def update_method_factory(instance: AnimatedNode, bound_update):  # noqa: ANN001 ANN202
            @wraps(bound_update)
            def new_update_method(delta: float) -> None:
                bound_update(delta)  # TODO: swap order will fix rendering??
                instance._wrapped_update_animated(delta)

            return new_update_method

        instance.update = update_method_factory(instance, instance.update)  # type: ignore
        return instance  # type: ignore

    animations: AnimationSet
    current_animation: Animation | None = None
    repeat: bool = False
    is_playing: bool = False
    _frame_index: int = 0
    _playback_direction: PlaybackDirection = PlaybackDirection.FORAWRD
    _is_on_last_frame: bool = False

    def with_animations(self, /, **animations: Animation) -> Self:
        # NOTE: additive
        for animation_name, animation in animations.items():
            setattr(self.animations, animation_name, animation)
        return self

    def with_animation(
        self,
        animation_name: str,
        animation: Animation,
        /,
    ) -> Self:
        # NOTE: additive
        self.add_animation(animation_name, animation)
        return self

    def with_repeat(self, state: bool = True, /) -> Self:
        self.repeat = state
        return self

    def add_animation(
        self,
        animation_name: str,
        animation: Animation,
        /,
    ) -> None:
        setattr(self.animations, animation_name, animation)

    def play(self, animation_name: str, /) -> None:
        if not hasattr(self.animations, animation_name):
            raise ValueError(f"animation not found: '{animation_name}'")
        self.current_animation = getattr(self.animations, animation_name)
        self.is_playing = True
        self._is_on_last_frame = False
        self._playback_direction = PlaybackDirection.FORAWRD
        self._frame_index = 0
        # the actual logic of playing the animation
        # is handled in `_wrapped_update_animated`

    def play_backwards(self, animation_name: str, /) -> None:
        if not hasattr(self.animations, animation_name):
            raise ValueError(f"animation not found: '{animation_name}'")
        self.current_animation = getattr(self.animations, animation_name)
        assert isinstance(self.current_animation, Animation)
        self.is_playing = True
        self._is_on_last_frame = False
        self._playback_direction = PlaybackDirection.BACKWARD
        self._frame_index = len(self.current_animation.frames) - 1
        # the actual logic of playing the animation
        # is handled in `_wrapped_update_animated`

    def _wrapped_update_animated(self, _delta: float) -> None:
        if self.current_animation is None:
            self.is_playing = False
            return

        self.texture = self.current_animation.frames[self._frame_index]
        frame_count = len(self.current_animation.frames)
        # using `min` and `max` instead of `clamp`
        # for better linting (`int` instead of `int | float`)
        index_change = 1 if self._playback_direction is PlaybackDirection.FORAWRD else -1
        self._frame_index = min(
            frame_count - 1,
            max(
                0,
                self._frame_index + index_change,
            ),
        )

        if self._is_on_last_frame:
            if self.repeat:
                self._frame_index = 0
                self._is_on_last_frame = False
            else:
                self.is_playing = False

        last_index = (
            frame_count - 1
            if self._playback_direction is PlaybackDirection.FORAWRD
            else 0
        )
        # state variable to ensure last frame is shown
        if self._frame_index == last_index:
            self._is_on_last_frame = True
